"""
   Copyright 2016 Net-Survey

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import logging
import multiprocessing
from multiprocessing.connection import Client, SocketListener
from oslo_config import cfg
from oslo_config import types
from platform import node
from Queue import Queue
from scanner import NetworkScanner
import socket
import sys
import time

SecondsType = types.Integer(1, 5000)
IntType = types.Integer(1, 50)

netsurvey_client_opts = [
    cfg.Opt('listener_timeout',
            type=SecondsType,
            default=900,
            help='Number of seconds to wait for client connections.'),
    cfg.Opt('timeout_wait',
            type=SecondsType,
            default=30,
            help='Number of seconds to sleep after timeout.'),
    cfg.Opt('scan_timeout',
            type=SecondsType,
            default=2,
            help='Number of seconds timeout for scan.'),
    cfg.StrOpt('scan_ports',
               default='1-1024',
               help='Ports to scan, i.e. 1-1024,9000,9696'),
    cfg.StrOpt('critical_ports',
               help='Ports neighboring hosts should be able to reach, and if '
                    'not, alert. This port must be in the scan_ports range.'),
    cfg.StrOpt('fake_ports',
               help='Ports this host should open a listener on, usually to '
                    'facilitate testing of critical ports where a service '
                    'is not actively running.'),
    cfg.Opt('concurrent_scans',
            type=IntType,
            default=1,
            help='Number of concurrent scans.'),
]


class NSClient(multiprocessing.Process):
    """Client process for net-survey."""

    UPDATING = 'updating'
    WAITING_DISTRIBUTION = 'awaiting_distribution'
    DISCOVERY = 'discovery'
    SEND_RESULTS = 'send_results'
    FINISHED = 'finished'
    SLEEPING = 'sleeping'
    SCAN_SUCCESSFUL = 'success'
    SCAN_FAILED = 'failed'
    START_SCAN = 'start_scan'

    def __init__(self, conf, interfaces):
        multiprocessing.Process.__init__(self, name='netsurvey')
        self._conf = conf
        self._interfaces = interfaces
        self._register_client_opts(self._conf)
        self._hostname = node()
        self._initialize_data()
        self._client = None
        self._server = None

    def _initialize_data(self):
        """Initialize Data & Cleanup."""
        self._networks = self._interfaces.get_host_networks()
        self._initialize_networks()
        self._state = self.SLEEPING
        self._client_data = {}
        self._result_data = {
            self._hostname: {}
        }
        self._ips_to_networks = {}
        self._ips_to_hostname = {}

    def _register_client_opts(self, conf):
        """Register server options.

        :param conf: a ConfigOpts instance
        :raises: DuplicateOptError, ArgsAlreadyParsedError
        """
        conf.register_opts(netsurvey_client_opts, group='client')

    def run(self):
        """Main run process for Client."""
        logging.info('Starting Client Process: '
                     '%s with pid %s' % (self.name, self.pid))
        while True:
            # Updating State
            if self._state == self.UPDATING and self._client is None:
                if self._send_data_to_server(self._networks):
                    logging.info('Networks sent: '
                                 'validated. Moving to %s.' %
                                 self.WAITING_DISTRIBUTION.upper())
                    self._state = self.WAITING_DISTRIBUTION
                else:
                    logging.debug('Sleeping for %s Seconds.' %
                                  self._conf.client.timeout_wait)
                    time.sleep(self._conf.client.timeout_wait)
            # Awaiting Distribution State
            elif self._state == self.WAITING_DISTRIBUTION:
                _received, _client_data = self._receive_data_from_server()
                if _received:
                    self._client_data.update(_client_data)
                    logging.info('Finished receiving distributed data. '
                                 'Moving to %s.' % self.DISCOVERY.upper())
                    self._state = self.DISCOVERY
                else:
                    logging.error('Waited a long time for the server to '
                                  'distribute data, closing and waiting.')
                    logging.debug('Sleeping for %s seconds.' %
                                  self._conf.client.timeout_wait)
                    time.sleep(self._conf.client.timeout_wait)
            # Discovery State
            elif self._state == self.DISCOVERY:
                _ports_to_scan = \
                    self._parse_ports(self._conf.client.scan_ports)
                _ips_to_scan = self._prepare_thread_ips(self._client_data)
                logging.debug('Ready to perform scans against ips: '
                              '%s' % _ips_to_scan)
                logging.debug('Scanning ports: %s' % _ports_to_scan)
                if not len(_ips_to_scan) > 0:
                    logging.warning('No neighboring hosts to scan. Shutting '
                                    'down process.')
                    self._terminate()
                else:
                    self._scan_ips(_ips_to_scan, _ports_to_scan,
                                   self._conf.client.concurrent_scans)
                    logging.debug('Scan Complete! Moving to %s.' %
                                  self.SEND_RESULTS.upper())
                    self._state = self.SEND_RESULTS
            elif self._state == self.SEND_RESULTS:
                logging.debug('Sending Results.')
                if self._send_data_to_server(self._result_data):
                    logging.info('Results sent: '
                                 'validated. Moving to %s.' %
                                 self.FINISHED.upper())
                    self._state = self.FINISHED
                else:
                    logging.debug('Sleeping for %s Seconds.' %
                                  self._conf.client.timeout_wait)
                    time.sleep(self._conf.client.timeout_wait)
            elif self._state == self.FINISHED:
                logging.info('Finished Scanning.')
                self._state = self.SLEEPING
            elif self._state == self.SLEEPING:
                _received, _command = self._receive_data_from_server()
                if _received and self.START_SCAN in str(_command):
                    logging.info('Received command from server to scan. '
                                 'Moving to %s.' % self.UPDATING.upper())
                    self._initialize_data()
                    self._state = self.UPDATING
                else:
                    logging.info('Waited a long time for the server to '
                                 'give a command, closing and waiting.')
                    logging.debug('Sleeping for %s seconds.' %
                                  self._conf.client.timeout_wait)
                    time.sleep(self._conf.client.timeout_wait)

    def _terminate(self):
        """Terminate method for process.

        This will cleanup client connection and terminate the process.
        """
        if self._server is not None:
            logging.info('Shutting Down Listener: '
                         '%s:%s' % (self._conf.bind_host,
                                    self._conf.bind_port))
            self._server.close()
        if self._client is not None:
            logging.info('Shutting Down Client: '
                         '%s:%s' % (self._conf.server_ip,
                                    self._conf.server_port))
            self._client.close()
        logging.info('Stopping Client Process: '
                     '%s with pid %s' % (self.name, self.pid))
        sys.exit(0)

    def terminate(self, sig, func=None):
        """Terminate method for process.

        This will cleanup client connection and terminate the process.

        :param sig: signal object
        :param func: function object
        """
        self._terminate()

    def _initialize_networks(self):
        """Adds results key & populates critical ports."""
        _critical_ports = {
            'critical_ports':
                self._parse_ports(self._conf.client.critical_ports)
        }
        for _network in self._networks[self._hostname]:
            self._networks[self._hostname][_network]['results'] = {}
            self._networks[self._hostname][_network].update(_critical_ports)

    def _parse_ports(self, ports):
        """Parse a config string of ports to scan.

        The config param is only allow '-' and ',' as separators, but will
        convert the string to an array of ports to scan.

        :param ports: config str of ports to scan
        :raises: TypeError

        :returns list of ports to scan from string of ports
        """
        _scan_ports = []

        if ports is not None:
            _str_ports_array = ports.split(',')

            for str_port in _str_ports_array:
                if '-' in str_port:
                    _start_port = int(str_port.split('-')[0])
                    _end_port = int(str_port.split('-')[-1])
                    for _port in range(_start_port, _end_port):
                        _scan_ports.append(_port)
                else:
                    _scan_ports.append(int(str_port))

        return _scan_ports

    def _prepare_thread_ips(self, client_data):
        """Prepare a list of ips this host needs to scan

        Scan through our own networks that we have connected and grab a list
        of neighboring hosts who have the same network.  If the host in that
        list does not equal our own hostname, go ahead and append the IP
        Addresses to a list of ips available for threads to scan.

        :param client_data: dict of prepared data to perform scans against
        :returns list of ips to scan
        """
        _scan_ips = []
        for _network in client_data['_host_detail'][self._hostname].keys():
            for _host in client_data['_networks'][_network]:
                if _host != self._hostname:
                    ips = client_data['_host_detail'][_host][_network]['ips']
                    if len(ips) > 0:
                        for ip in ips:
                            _scan_ips.append(ip)
                            # Need to keep a relationship of ips to networks
                            # and host for posting results.
                            self._ips_to_networks[ip] = _network
                            self._ips_to_hostname[ip] = _host

        return _scan_ips

    def _scan_ips(self, _ips, _ports, _concurrent_scans):
        """Perform scan on neighboring IP Addresses

        Setup queues and threads per IP and consume the concurrent_threads
        configuration to determine how many will be scanning at one time.

        State list will keep track of all threads and their two states:
        scanning and hold_scan. It's important to pause after sending then
        scan command to the queue, as it will take at most 1 second for the
        thread to consume it.

        :param _ips: list of ips to scan
        :param _ports: list of ports to scan
        :param _concurrent_scans: int number of concurrent scans
        """
        _num_threads = len(_ips)
        _queues = [x for x in range(0, _num_threads)]
        _threads = [x for x in range(0, _num_threads)]
        _start_events = [x for x in range(0, _num_threads)]
        _stop_events = [x for x in range(0, _num_threads)]
        _ip_scanned_ports = {ip: [] for ip in _ips}

        # Setup queues, threads, and events
        for x in range(0, _num_threads):
            _queues[x] = Queue()
            _start_events[x] = multiprocessing.Event()
            _stop_events[x] = multiprocessing.Event()
            # Get the IP's critical ports and add them to the scan list
            _ip_scanned_ports[_ips[x]] = _ports
            _critical_ports = self._client_data['_host_detail'][
                self._ips_to_hostname[_ips[x]]][
                self._ips_to_networks[_ips[x]]][
                'critical_ports']
            for port in _critical_ports:
                if port not in _ports:
                    _ip_scanned_ports[_ips[x]].append(port)
            _threads[x] = NetworkScanner(_ips[x], _ip_scanned_ports[_ips[x]],
                                         _queues[x], _start_events[x],
                                         _stop_events[x], self._conf)

        while True:
            # Get the current number of scanning threads
            _num_scanning = 0
            for event in _start_events:
                if event.is_set():
                    _num_scanning += 1

            if _num_scanning < _concurrent_scans:
                _need_to_scan = _concurrent_scans - _num_scanning
                for event in enumerate(_start_events):
                    if not event[1].is_set():
                        logging.debug('Informing thread to scan ip %s.' %
                                      _ips[event[0]])
                        _threads[event[0]].start()
                        event[1].set()
                        _need_to_scan -= 1
                    if _need_to_scan == 0:
                        break

            for event in enumerate(_stop_events):
                if event[1].is_set() and not _queues[event[0]].empty():
                    _thread_result = _queues[event[0]].get()
                    _index = event[0]
                    logging.debug('Received data: %s' % _thread_result)
                    if self.SCAN_SUCCESSFUL in str(_thread_result) \
                            or self.SCAN_FAILED in str(_thread_result):
                        self._receive_scan_data(_ips[_index],
                                                _queues[_index],
                                                _thread_result,
                                                _ip_scanned_ports[
                                                    _ips[_index]])
                        del _ips[_index]
                        del _queues[_index]
                        del _start_events[_index]
                        del _stop_events[_index]
                        del _threads[_index]
            if len(_queues) == 0:
                break

    def _receive_scan_data(self, ip, queue, result, scanned_ports):
        """Receive result data from scan.

        This will check the queue if result is successful and grab the result
        dict.

        :param ip: IP Address from _ips
        :param queue; Queue object from _queues
        :param result: _thread_result, successful scan or not
        :param scanned_ports: list of scanned ports
        """
        _scan_results = []
        _failed_critical_ports = []
        if self.SCAN_SUCCESSFUL in str(result) and not queue.empty():
            logging.debug('Received successful scan: Host(%s) Data(%s).' %
                          (ip, result))
            _scan_results = queue.get()
            _reason = 'reachable'
            logging.debug('Received results: Host(%s) Data(%s).' %
                          (ip, _scan_results))

            # Check our critical ports for the scanned ip
            _critical_ports = self._client_data['_host_detail'][
                self._ips_to_hostname[ip]][
                self._ips_to_networks[ip]]['critical_ports']
            if self.SCAN_FAILED not in str(result):
                for port in _critical_ports:
                    if port not in list(_scan_results):
                        _failed_critical_ports.append(port)
                        result = self.SCAN_FAILED
                        _reason = 'critical_ports'
        else:
            _reason = 'unreachable'
            logging.debug('Received failed scan: '
                          'Host(%s) Data(%s).' %
                          (ip, result))

        _result = {}
        _result[ip] = {
            'result': result,
            'reason': _reason,
            'open_ports': _scan_results,
            'host': self._ips_to_hostname[ip],
            'scanned_ports': scanned_ports,
            'failed_critical_ports': _failed_critical_ports
        }

        logging.debug('Final results: %s' % _result)
        self._client_data['_host_detail'][
            self._hostname
        ][
            self._ips_to_networks[ip]
        ][
            'results'].update(_result)
        self._result_data[self._hostname] = \
            self._client_data['_host_detail'][self._hostname]

    def _send_data_to_server(self, data):
        """Send data to server.

        This will send the data to the server, i.e. networks or
        results.

        :returns True/False if data is validated/received from server
        """
        try:
            logging.debug('Sending data: (%s:%s) - '
                          '%s' % (self._conf.server_ip,
                                  self._conf.server_port,
                                  data))
            self._client = Client((self._conf.server_ip,
                                   self._conf.server_port))
            self._client.send(data)
            response = self._client.recv()
            if str(response) == 'validated':
                self._client.close()
                return True
        except Exception as e:
            logging.error('Unable to connect to server: '
                          '%s:%s - %s' % (self._conf.server_ip,
                                          self._conf.server_port,
                                          str(e)))
            return False
        finally:
            self._client = None

    def _receive_data_from_server(self):
        """Receive data from server.

        This method will retrieve the distributed data from the server.

        :returns bool(if data received), _client_data(dict of data received)
        """
        logging.info('Starting Listener: '
                     '%s:%s' % (self._conf.bind_host,
                                self._conf.bind_port))
        self._server = SocketListener((self._conf.bind_host,
                                       self._conf.bind_port), 'AF_INET')
        _recieved = False
        _client_data = None
        while True:
            # Bypass Listener and use SocketListener to set timeout
            self._server._socket.settimeout(self._conf.client.listener_timeout)
            try:
                connection = self._server.accept()
            except socket.timeout:
                # Send us back to run() to communicate the timeout.
                break
            while True:
                try:
                    _client_data = connection.recv()
                    logging.debug('Distributed data received from server: '
                                  '%s - Data: %s.' %
                                  (self._server._last_accepted[0],
                                   _client_data))
                    connection.send('validated')
                    _recieved = True
                except EOFError:
                    break
            connection.close()
            if _recieved:
                break
        self._server.close()
        self._server = None
        return _recieved, _client_data



"""
   Copyright 2016 Net-Survey

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from BaseHTTPServer import HTTPServer
from BaseHTTPServer import BaseHTTPRequestHandler
import json
import logging
import multiprocessing
import re
import redis


class APIServer(multiprocessing.Process):

    def __init__(self, conf, redis_pool):
        multiprocessing.Process.__init__(self, name='apiserver',)
        self._conf = conf
        self._httpd_server = APIHTTPServer((self._conf.bind_host,
                                            self._conf.server_api_port),
                                           MainRequestHandler)
        self._httpd_server.set_redis_server(redis_pool)

    def run(self):
        logging.info('Starting API Process: '
                     '%s with pid %s' % (self.name, self.pid))
        try:
            logging.info('Starting API Server: %s:%s.' %
                         (self._conf.bind_host, self._conf.server_api_port))
            self._httpd_server.serve_forever()
        except Exception as e:
            logging.error('API Server Shutdown: %s:%s. Exception: %s' %
                          (self._conf.bind_host,
                           self._conf.server_api_port,
                           str(e)))
            self._terminate()

    def _terminate(self):
        """Terminate method for process.

        This will cleanup client connection and terminate the process.
        """
        logging.info('Shutting Down API Server: %s:%s' %
                     (self._conf.bind_host, self._conf.server_api_port))
        self._httpd_server.socket.close()
        logging.info('Stopping API Process: '
                     '%s with pid %s' % (self.name, self.pid))

    def terminate(self, sig, func=None):
        """Terminate method for process.

        This will cleanup client connection and terminate the process.

        :param sig: signal object
        :param func: function object
        """
        self._terminate()


class APIHTTPServer(HTTPServer):
    """HTTPServer Class to Set Queue and Event."""

    def set_redis_server(self, redis_pool):
        """Set Queue and Event Instances.

        :param redis_ip: redis server IP Address
        :param redis_port: redis server port
        """
        self.redis_pool = redis_pool


class Controller(object):
    """Controller Ref to Base Class."""

    def __init__(self, server):
        self.__server = server

    @property
    def server(self):
        """Ref to Base Class."""
        return self.__server


class Router(object):
    """Base Router."""

    def __init__(self, server):
        """Base Router.

        :param server: HTTPServer instance
        """
        self.__routes = []
        self.__server = server

    def addRoute(self, regexp, controller, action):
        """Add Route to Controller.

        :param regexp: regexp match on route
        :param controller: controller object
        :param action: controller action
        """
        self.__routes.append({'regexp': regexp,
                              'controller': controller,
                              'action': action})

    def route(self, path):
        """Check Route and Execute.

        :param path: URI Path
        """
        for route in self.__routes:
            if re.search(route['regexp'], path):
                cls = globals()[route['controller']]
                func = cls.__dict__[route['action']]
                obj = cls(self.__server)
                apply(func, (obj, ))
                return

        # 404 Not Found
        self.__server.send_response(404)
        self.__server.end_headers()


class MainRequestHandler(BaseHTTPRequestHandler):
    """Main Request Handler."""

    def __init__(self, request, client_address, server):

        self._redis_pool = server.redis_pool

        get_routes = [
            {'regexp': r'^/$',
             'controller': 'MainController',
             'action': 'indexAction'}
        ]

        post_routes = [
            {'regexp': r'^/start$',
             'controller': 'MainController',
             'action': 'scanAction'}
        ]

        self.__get_router = Router(self)
        self.__post_router = Router(self)

        for route in get_routes:
            self.__get_router.addRoute(route['regexp'],
                                       route['controller'],
                                       route['action'])

        for route in post_routes:
            self.__post_router.addRoute(route['regexp'],
                                        route['controller'],
                                        route['action'])

        BaseHTTPRequestHandler.__init__(self, request, client_address, server)

    def do_GET(self):
        """Handle GETs."""
        self.__get_router.route(self.path)

    def do_POST(self):
        """Handle POSTs."""
        self.__post_router.route(self.path)


class MainController(Controller):
    """Main Controller."""

    START_SCAN = 'start_scan'

    def __init__(self, server):
        """Main Controller.

        :param server: HTTPServer instance
        """
        Controller.__init__(self, server)

    def _set_redis_action(self, action):
        """Sets the action on Redis action key.

        :param action: string action to set
        """
        if self.server._redis_pool is not None:
            try:
                r = redis.Redis(connection_pool=self.server._redis_pool)
                r.set('action', action)
            except Exception as e:
                logging.error('Unable to set action in Redis: %s' % str(e))
        else:
            logging.error('Redis connection not setup. Unable to set action.')

    def _get_redis_data(self):
        """Sets the action on Redis action key.

        :returns json object of scan data
        """
        if self.server._redis_pool is not None:
            try:
                r = redis.Redis(connection_pool=self.server._redis_pool)
                return json.loads(r.get('scan_data'))
            except Exception as e:
                logging.error('Unable to set action in Redis: %s' % str(e))
        else:
            logging.error('Redis connection not setup. Unable to set action.')

    def indexAction(self):
        """Handle GET to Index.

        This will set the event flag and wait for queue data.
        """
        _data = {}

        try:
            _data = self._get_redis_data()
        except Exception as e:
            logging.error('Unable to retrieve Redis scan data: %s' % str(e))

        self.server.send_response(200)
        self.server.send_header('Content-type', 'application/json')
        self.server.end_headers()
        # Write our data and clear the flag
        self.server.wfile.write(json.dumps(_data))

    def scanAction(self):
        """Handle POST to Index.

        This will start a scan.
        """
        _data = "OK"

        self._set_redis_action(self.START_SCAN)
        self.server.send_response(201)
        self.server.wfile.write(_data)



"""
   Copyright 2016 Net-Survey

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import json
import logging
import multiprocessing
from multiprocessing.connection import Client, SocketListener
from oslo_config import cfg
from oslo_config import types
import redis
import socket
import sys
import time

SecondsType = types.Integer(1, 5000)

netsurvey_server_opts = [
    cfg.Opt('listener_timeout',
            type=SecondsType,
            default=900,
            help='Number of seconds to wait for client connections.'),
    cfg.Opt('timeout_wait',
            type=SecondsType,
            default=30,
            help='Number of seconds to sleep after timeout.'),
    cfg.BoolOpt('scan_on_start',
                default=False,
                help='Start scan on server service start '
                     'True/False or yes/no.')
]


class NSServer(multiprocessing.Process):
    """Server process for net-survey."""

    LEARNING = 'learning'
    PROCESSING = 'processing'
    DISTRIBUTING = 'distributing'
    GATHERING = 'gathering'
    FINISHED = 'finished'
    SLEEPING = 'sleeping'
    START_SCAN = 'start_scan'

    def __init__(self, conf, redis_pool):
        multiprocessing.Process.__init__(self, name='netsurvey',)
        self._conf = conf
        self._register_server_opts(self._conf)
        self._server = None
        self._client = None
        self._redis_pool = redis_pool
        self._initialize_data()
        if bool(conf.server.scan_on_start):
            self._state = self.START_SCAN

    def _initialize_data(self):
        """Initialize Data & Cleanup."""
        self._state = self.SLEEPING
        self._client_data = {}
        self._client_data['_state'] = self.SLEEPING
        self._client_data['_host_detail'] = {}
        self._client_data['_networks'] = {}
        self._client_data['_clients'] = {}
        self._clients = self._populate_clients()
        self._update_redis_data()
        logging.debug('Populated Inventory: %s' % self._clients)

    def _update_redis_data(self):
        """Update Redis Data."""
        if self._redis_pool is not None:
            try:
                r = redis.Redis(connection_pool=self._redis_pool)
                _json_data = json.dumps(self._client_data, ensure_ascii=False)
                r.set('scan_data', _json_data)
            except Exception as e:
                logging.error('Unable to update data in Redis: %s' % str(e))
        else:
            logging.error('Redis connection not setup. Unable to update data.')

    def _get_redis_action(self):
        """Get Redis Action.

        :returns bool if scan should start
        """
        _scan = False

        if self._redis_pool is not None:
            try:
                r = redis.Redis(connection_pool=self._redis_pool)
                if self.START_SCAN in str(r.get('action')):
                    _scan = True
                    r.set('action', 'clear')
            except Exception as e:
                logging.error('Unable to get action in Redis: %s' % str(e))
        else:
            logging.error('Redis connection not setup. Unable to get action.')

        return _scan

    def _register_server_opts(self, conf):
        """Register server options.

        :param conf: a ConfigOpts instance
        :raises: DuplicateOptError, ArgsAlreadyParsedError
        """
        conf.register_opts(netsurvey_server_opts, group='server')

    def run(self):
        """Main run process for Server."""

        # Remaining clients to distribute data to.
        _remaining_clients = []
        logging.info('Starting Server Process: '
                     '%s with pid %s' % (self.name, self.pid))
        while True:
            # Learning State
            if self._state == self.LEARNING and self._server is None:
                _complete = self._receive_data_from_clients()
                if _complete:
                    logging.info('Finished receiving client data. '
                                 'Moving to %s.' % self.PROCESSING.upper())
                    self._update_redis_data()
                    self._state = self.PROCESSING
                    self._client_data['_state'] = self._state
                else:
                    logging.error('Waited a long time for all clients to '
                                  'report network data but timed out waiting. '
                                  'Check hosts: %s' %
                                  [ip for ip in self._clients.keys()
                                   if self._clients[ip] is None])
                    logging.debug('Sleeping for %s seconds.' %
                                  self._conf.server.timeout_wait)
                    time.sleep(self._conf.server.timeout_wait)

            # Processing State
            elif self._state == self.PROCESSING:
                self._prepare_client_data()
                logging.debug('Client data prepared: %s' % self._client_data)
                logging.info('Moving to %s.' % self.DISTRIBUTING.upper())
                self._update_redis_data()
                self._state = self.DISTRIBUTING
                self._client_data['_state'] = self._state

            # Distribution State
            elif self._state == self.DISTRIBUTING:
                _distributed, _remaining_clients = \
                    self._distribute_data(_remaining_clients,
                                          self._client_data)
                if _distributed:
                    logging.debug('Client data distributed: '
                                  '%s' % self._client_data)
                    logging.info('Moving to %s.' % self.GATHERING.upper())
                    self._state = self.GATHERING
                    self._client_data['_state'] = self._state
                    # Setup clients dict for tracking again
                    self._clients = self._populate_clients()
                else:
                    logging.critical('Unable to distribute to clients. Trying '
                                     'again in %s seconds. Clients in error: '
                                     '%s' % (self._conf.server.timeout_wait,
                                             _remaining_clients))
                    time.sleep(self._conf.server.timeout_wait)
                self._update_redis_data()

            # Gathering Results State
            elif self._state == self.GATHERING:
                _complete = self._receive_data_from_clients()
                if _complete:
                    logging.info('Finished receiving client data. '
                                 'Moving to %s.' % self.FINISHED.upper())
                    self._state = self.FINISHED
                    self._client_data['_state'] = self._state
                else:
                    logging.error('Waited a long time for all clients to '
                                  'report network data but timed out waiting. '
                                  'Check hosts: %s' %
                                  [ip for ip in self._clients.keys()
                                   if self._clients[ip] is None])
                    logging.debug('Sleeping for %s seconds.' %
                                  self._conf.server.timeout_wait)
                    time.sleep(self._conf.server.timeout_wait)
                self._update_redis_data()

            # Finished State
            elif self._state == self.FINISHED:
                logging.info('Finished Scanning.')
                logging.debug('Final client result data: %s' %
                              self._client_data)
                self._update_redis_data()
                self._state = self.SLEEPING
                self._client_data['_state'] = self._state

            # Start Scan
            elif self._state == self.START_SCAN:
                _distributed, _remaining_clients = \
                    self._distribute_data(_remaining_clients,
                                          self.START_SCAN)
                if _distributed:
                    logging.debug('Client Initialized: '
                                  '%s' % self.START_SCAN.upper())
                    logging.info('Moving to %s.' % self.LEARNING.upper())
                    self._initialize_data()
                    self._state = self.LEARNING
                    self._client_data['_state'] = self._state
                else:
                    logging.critical('Unable to initialize clients. Trying '
                                     'again in %s seconds. Clients in error: '
                                     '%s' % (self._conf.server.timeout_wait,
                                             _remaining_clients))
                    time.sleep(self._conf.server.timeout_wait)
            elif self._state == self.SLEEPING:
                # Check scan event
                logging.debug('Sleeping for 5 seconds. State: %s.' %
                              self._state.upper())
                time.sleep(5)
                self._check_scan_event()

    def _terminate(self):
        """Terminate method for process.

        This will cleanup client connection and terminate the process.
        """
        if self._server is not None:
            logging.info('Shutting Down Listener: '
                         '%s:%s' % (self._conf.bind_host,
                                    self._conf.bind_port))
            self._server.close()
        if self._client is not None:
            logging.info('Shutting Down Client.')
            self._client.close()
        logging.info('Stopping Server Process: '
                     '%s with pid %s' % (self.name, self.pid))
        sys.exit(0)

    def terminate(self, sig, func=None):
        """Terminate method for process.

        This will cleanup client connection and terminate the process.

        :param sig: signal object
        :param func: function object
        """
        self._terminate()

    def _receive_data_from_clients(self):
        """Receive data from clients.

        This method receives data from clients and returns true if data
        has been received.  If so, decision to transition to another state
        is done within the run() method.

        One check is made here, which is to check if data has been received
        from all clients in the states LEARNING and GATHERING.

        :returns bool(if data received), _client_data(dict of data received)
        """
        logging.info('Starting Listener: '
                     '%s:%s with Timeout: %s seconds' %
                     (self._conf.bind_host, self._conf.bind_port,
                      self._conf.server.listener_timeout))
        self._server = SocketListener((self._conf.bind_host,
                                       self._conf.bind_port), 'AF_INET')
        _complete = False
        while True:
            # Bypass Listener and use SocketListener to set timeout
            self._server._socket.settimeout(self._conf.server.listener_timeout)
            try:
                connection = self._server.accept()
            except socket.timeout:
                # Send us back to run() to communicate the timeout.
                break
            while True:
                try:
                    _recv_client_data = connection.recv()
                    logging.debug('Client data received from: '
                                  '%s - Data: %s.' %
                                  (self._server._last_accepted[0],
                                   _recv_client_data))
                    # We update the _clients dict, adding the hostname
                    # to the ip key, if does not exist, we traverse the
                    # client_data looking for the ip.
                    self._update_clients(self._server._last_accepted[0],
                                         _recv_client_data)
                    self._client_data['_host_detail'].update(_recv_client_data)
                    connection.send('validated')
                    self._update_redis_data()
                except EOFError:
                    break
            connection.close()
            # Check the _clients dict to make sure we received all
            # client information.
            if self._check_clients():
                _complete = True
                break
        self._server.close()
        self._server = None
        return _complete

    def _populate_clients(self):
        """Populate the _clients dict.

        :returns: dict of client ips in inventory file with empty hostname
        """
        client_file = open(self._conf.inventory_file, 'r')
        return {line.strip(): None
                for line in client_file.readlines()
                if '[' not in line}

    def _update_clients(self, host_ip, client_data):
        """Updates _clients dict with hostname of client."""
        try:
            for key in client_data.keys():
                self._clients[host_ip] = key
                logging.debug('Updated _client dict with hostname: %s.' % key)
        except KeyError as e:
            logging.debug('Unable to locate client ip in _clients dict. '
                          'We will now traverse client_data to find the '
                          'inventory ip address. %s' % str(e))

    def _check_clients(self):
        """Traverses the _clients dict to check completion.

        :returns: True/False if all client information has been received
        """
        _update_complete = True
        for value in self._clients.values():
            if value is None:
                _update_complete = False
        return _update_complete

    def _prepare_client_data(self):
        """Prepare data to be sent to the client for testing.

        A dict of networks and hostnames which have ips on the network are
        compiled as a key to retrieve ips to test against.

        { '_clients':
                'hostname: 'IP'
          '_networks':
                '10.0.0.0/24': ['host1', 'host2']
          '_host_detail':
                'hostname':
                    '10.0.0.1/24':
                        'ips': ['10.0.0.10' ,'10.0.0.2']
        }
        """

        # Loop through our networks for each host, determine if the network
        # has been added in the _networks key, if so just append the hostname
        # to the array.  If not, create they key and append.
        for _host in self._client_data['_host_detail'].keys():
            for _network in self._client_data['_host_detail'][_host].keys():
                if _network not in self._client_data['_networks'].keys():
                    self._client_data['_networks'][_network] = []
                self._client_data['_networks'][_network].append(_host)

        # Create a reverse dict of _clients for loopkup in the above network
        # reference.
        logging.debug('Processing clients: %s.' % self._clients)
        self._client_data['_clients'] = {self._clients[_host_ip]: _host_ip
                                         for _host_ip
                                         in self._clients.keys()}

    def _distribute_data(self, remaining_clients, data):
        """Iterates over a list of clients to distribute data to.

        List is compiled of unreachable clients for use on a retry.  If there
        are hosts in remaining_clients, the _clients dict is no longer used.

        :param remaining_clients: list of unreachable clients to
        distribute test data to.
        :param data: data to send to the client

        :returns: True/False if all client information has been sent, and
        a list of unreachable clients to use for a retry.
        """
        _distribute_to = []
        _errored_clients = []
        _distributed = True
        if len(remaining_clients) > 0:
            _distribute_to = remaining_clients
        else:
            _distribute_to = [_ip for _ip in self._clients.keys()]

        for _host_ip in _distribute_to:
            if not self._send_data_to_client(_host_ip,
                                             self._conf.bind_port,
                                             data):
                _errored_clients.append(_host_ip)
                _distributed = False

        return _distributed, _errored_clients

    def _send_data_to_client(self, client_ip, client_port, data):
        """Send data to server.

        This will send the finalized test data to the client.

        :returns True/False if data is validated/received from client
        """
        try:
            logging.debug('Sending data: (%s:%s) - '
                          '%s' % (client_ip, client_port, data))
            self._client = Client((client_ip, client_port))
            self._client.send(data)
            response = self._client.recv()
            if str(response) == 'validated':
                self._client.close()
                return True
        except Exception as e:
            logging.error('Unable to connect to client: '
                          '%s:%s - %s' % (client_ip,
                                          client_port,
                                          str(e)))
            return False
        finally:
            self._client = None

    def _check_scan_event(self):
        """Checks the Redis Action.

        This will check the action key in Redis and if set will start a scan
        if in the SLEEPING state.
        """
        if self._get_redis_action():
            logging.info('Recieved scan event.')
            self._state = self.START_SCAN
            self._client_data['_state'] = self._state



"""
   Copyright 2016 Net-Survey

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from client.client import NSClient
from server.api import APIServer
from interfaces import NetworkInterfaces
import logging
from oslo_config import cfg
from oslo_config import types
import pkg_resources
import redis
import signal
import sys
from server.server import NSServer

PortType = types.Integer(1, 65535)

_cli_opts = [
    cfg.StrOpt('mode',
               default='client',
               help='Survey mode: server, api, or client.'),
    cfg.StrOpt('bind_host',
               default='0.0.0.0',
               help='IP Address to listen on.'),
    cfg.Opt('bind_port',
            type=PortType,
            default=4000,
            help='Port to listen on.'),
    cfg.StrOpt('server_ip',
               default='127.0.0.1',
               help='IP Address of survey master server.'),
    cfg.Opt('server_port',
            type=PortType,
            default=4000,
            help='Port to connect to the server on.'),
    cfg.Opt('server_api_port',
            type=PortType,
            default=8000,
            help='HTTP API Port.'),
    cfg.IPOpt('redis_server_ip',
              default='127.0.0.1',
              help='Redis server IP Address.'),
    cfg.Opt('redis_server_port',
            type=PortType,
            default=6379,
            help='Number of seconds to wait for client connections.'),
    cfg.StrOpt('inventory_file',
               default='/etc/net-survey/inventory.ini',
               help='Inventory file used for client connections.'),
    cfg.BoolOpt('use_log_file',
                default=True,
                help='Use log file for output True/False or yes/no.'),
    cfg.StrOpt('log_file',
               default='/var/log/net-survey/survey.log',
               help='Log file location.'),
    cfg.StrOpt('log_level',
               default='INFO',
               help='Logging level: DEBUG, INFO, WARNING, ERROR, etc.'),
]


def register_cli_opts(conf):
    """Register net-survey cli options.

    :param conf: a ConfigOpts instance
    :raises: DuplicateOptError, ArgsAlreadyParsedError
    """
    conf.register_cli_opts(_cli_opts)


def _setup_redis(conf):
        """Connect to Redis Server.

        :param conf: a ConfigOpts instance
        :returns redis pool instance
        """
        _redis_pool = None

        try:
            _redis_pool = redis.ConnectionPool(host=conf.redis_server_ip,
                                               port=conf.redis_server_port,
                                               db=0)
            logging.info('Redis Connection Pool Setup: %s:%s' %
                         (conf.redis_server_ip,
                          conf.redis_server_port))
        except Exception as e:
            logging.error('Unable to setup Redis: %s' % str(e))

        return _redis_pool


def _setup_logging(conf):
    """Setup logging level and output file.

    :param conf: a ConfigOpts instance
    :raises: KeyError
    """
    _levels = {
        'CRITICAL': 50,
        'ERROR': 40,
        'WARNING': 30,
        'INFO': 20,
        'DEBUG': 10,
        'NOTSET': 0
    }

    try:
        _conf_level = _levels[conf.log_level.upper()]
        if bool(conf.use_log_file):
            logging.basicConfig(filename=conf.log_file, level=_conf_level)
        else:
            logging.basicConfig(level=_conf_level)
    except Exception as e:
        logging.basicConfig(level=_levels['DEBUG'])
        logging.error('Error setting log level: %s' % str(e))


def main(args=None):
    """The main function of net-survey."""
    version = pkg_resources.get_distribution('oslo.config').version
    conf = cfg.ConfigOpts()
    register_cli_opts(conf)
    conf(args, version=version)
    _setup_logging(conf)

    if conf.mode == 'server' or conf.mode == 'api' or conf.mode == 'mixed':
        # Setup our redis pool
        _redis = _setup_redis(conf)

        if conf.mode == 'server' or conf.mode == 'mixed':
            if not conf.inventory_file:
                logging.critical('Inventory file must be speicified when '
                                 'running in server mode.')
                sys.exit(1)

            # Start Net-Survey Server
            srv = NSServer(conf, _redis)
            signal.signal(signal.SIGTERM, srv.terminate)
            signal.signal(signal.SIGINT, srv.terminate)
            srv.start()

        if conf.mode == 'api' or conf.mode == 'mixed':
            # Start API server
            api_srv = APIServer(conf, _redis)
            signal.signal(signal.SIGTERM, api_srv.terminate)
            signal.signal(signal.SIGINT, api_srv.terminate)
            api_srv.start()
    else:
        cli = NSClient(conf, NetworkInterfaces())
        signal.signal(signal.SIGTERM, cli.terminate)
        signal.signal(signal.SIGINT, cli.terminate)
        cli.start()


if __name__ == '__main__':
    main()



"""
   Copyright 2016 Net-Survey

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import ipaddress
import logging
import netifaces as ni
import platform


class NetworkInterfaces(object):
    """Retrieves host interface information."""

    _ignore_interfaces = ['lo', 'lo0']
    _hostname = platform.node()
    _host_interfaces = {}
    _host_networks = {}
    _interfaces = []
    _interface_dict = {}

    def __init__(self):
        self._interfaces = self._scrub_interfaces(ni.interfaces())
        self._host_interfaces[self._hostname] = {}
        self._host_networks[self._hostname] = {}

    def _scrub_interfaces(self, _ints):
        """Scrub the interfaces.

        Removes unneeded interfaces from the interface list.

        :param _ints: a list of interfaces
        :returns: a list of scrubbed interfaces
        """
        for interface in self._ignore_interfaces:
            if interface in _ints:
                _ints.remove(interface)

        return _ints

    def _get_network(self, _ip, _netmask):
        """Get network for host interface ip.

        Uses ipaddress module to get the network for host interface ip.

        :param _ip: str object of ip address
        :param _netmask: str object of netmask
        :returns: IPv4Network object describing the network/cidr
        """
        _cidr = ipaddress.IPv4Network(u"0.0.0.0/%s" % _netmask).prefixlen
        _host_ip = ipaddress.ip_interface(u"%s/%s" % (_ip, _cidr))
        _network = _host_ip.network

        return _network

    def get_interfaces(self):
        """Populate the interface dict.

        Creates a dict with scrubbed interfaces and ips.
        {'interfaces':
            'eth0':
                'network': '10.0.0.0/24'
                'addresses': [{'addr': '10.0.0.10',
                               'netmask': '255.255.255.0',
                               'gateway': '10.0.0.1'}]
        }

        :returns: a dict of interfaces and ip addresses
        """
        for _interface in self._interfaces:
            _addresses = ni.ifaddresses(_interface)
            if ni.AF_INET in _addresses:
                self._interface_dict[_interface] = {}
                self._interface_dict[_interface]['addresses'] = \
                    _addresses[ni.AF_INET]
                # (Note: ski) addresses will contain a list of ips:
                # primary, secdonary and network will have the
                # calculated network/cidr
                _first_ip = _addresses[ni.AF_INET][0]['addr']
                _first_nm = _addresses[ni.AF_INET][0]['netmask']
                try:
                    self._interface_dict[_interface]['network'] = \
                        str(self._get_network(_first_ip, _first_nm))
                except (ipaddress.AddressValueError,
                        ipaddress.NetmaskValueError, OSError) as e:
                    logging.error("Error retrieving ip "
                                  "interface network: {}".format(str(e)))

        return self._interface_dict

    def get_host_interfaces(self):
        """Populate the host interfaces dict.

        Creates a dict with the hostname as the key.
        {'hostname':
            'interfaces':
                'eth0':
                    'network': '10.0.0.0/24'
                    'addresses': [{'addr': '10.0.0.10',
                                   'netmask': '255.255.255.0',
                                   'gateway': '10.0.0.1'}]
        }

        :returns: a dict of all host interfaces, ips, networks
        """
        self._host_interfaces[self._hostname]['interfaces'] = \
            self.get_interfaces()

        return self._host_interfaces

    def get_host_networks(self):
        """Populate the hsot networks dict.

        Creates a dict with the hostname as the key for all networks.
        {'hostname':
            '10.0.0.1/24':
                'ips': ['10.0.0.10' ,'10.0.0.2']
        }

        :returns: a dict of all host networks and ips on the network
        """
        _interfaces = self.get_host_interfaces()[self._hostname]['interfaces']

        for int in _interfaces:
            _network = _interfaces[int]['network']
            _addresses = _interfaces[int]['addresses']
            self._host_networks[self._hostname][_network] = {}
            self._host_networks[self._hostname][_network]['ips'] = []
            for address in _addresses:
                self._host_networks[self._hostname][_network]['ips']\
                    .append(address['addr'])

        return self._host_networks



"""
   Copyright 2016 Net-Survey

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import logging
import threading
import socket


class NetworkScanner(threading.Thread):
    """Client NetworkScanner thread for net-survey."""

    SCAN_SUCCESSFUL = 'success'
    SCAN_FAILED = 'failed'
    START_SCAN = 'start_scan'
    HOLD_SCAN = 'hold_scan'

    def __init__(self, server_ip, ports, queue, start_event, stop_event, conf):
        self._server_ip = server_ip
        self._ports = ports
        self._queue = queue
        self._start_event = start_event
        self._stop_event = stop_event
        self._conf = conf
        self._open_ports = []
        threading.Thread.__init__(self)

    def run(self):
        """Main run process for NetworkScanner."""

        _scan_result = self.SCAN_FAILED

        # Wait for start event
        self._start_event.wait()
        _scan_result, self._open_ports = \
            self.scan_host(self._server_ip, self._ports)

        # Notify stop event, status of scan, and data
        self._stop_event.set()
        self._queue.put(_scan_result)
        if _scan_result == self.SCAN_SUCCESSFUL:
            self._queue.put(self._open_ports)

    def scan_host(self, ip, ports):
        """Scan host thread.

        This method performs work on the neighboring ip and scans the ports
        passed to the class instance.

        :param ip: a valid ip address
        :param ports: a list of ports to scan

        :returns bool success, list of open ports
        """
        _success = self.SCAN_SUCCESSFUL
        _results = []

        for port in ports:
            try:
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.settimeout(self._conf.client.scan_timeout)
                logging.debug('Scanning host: %s:%s' % (self._server_ip, port))
                result = s.connect_ex((ip, port))
                if result == 0:
                    _results.append(port)
                s.close()
            except socket.gaierror:
                logging.error('Hostname could not be resolved: %s' %
                              ip)
                _success = self.SCAN_FAILED
                break
            except socket.error:
                logging.error('Cloud not connect to host: %s.' %
                              ip)
                _success = self.SCAN_FAILED
                break

        if len(_results) == 0:
            logging.error('No ports open, cannot determine if host is '
                          'reachable.')
            _success = self.SCAN_FAILED

        return _success, _results



"""
   Copyright 2016 Project Name

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import unittest


class BaseTests(unittest.TestCase):
    """Base unit tests. """

    def setUp(self):
        """Setup for base test. """
        pass

    def tearDown(self):
        """Teardown for base test. """

    def test_base_unit(self):
        """Base test. """
        pass



#!/usr/bin/python
"""
   Copyright 2016 Project Name

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from setuptools import setup, find_packages
from codecs import open
from os import path, makedirs

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Create /etc/net-survey
if not path.exists('/etc/net-survey'):
    makedirs('/etc/net-survey')

# Create /var/log/net-survey
if not path.exists('/var/log/net-survey'):
    makedirs('/var/log/net-survey')

# Setup requirements
with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='net-survey',

    version='0.0.4',

    description='Network Installation Survey',
    long_description=long_description,

    url='https://github.com/stajkowski/net-survey',

    author='Brian Stajkowski',
    author_email='stajkowski100@gmail.com',

    license='Apache',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    keywords='network survey',

    install_requires=required,

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    data_files=[('/etc/net-survey', ['etc/inventory.ini',
                                     'etc/netsurvey.config'])],

    entry_points={
        'console_scripts': [
            'net-survey=net_survey.survey:main',
        ],
    },
)
